from typing import Any

from .abstract import CTree
from .factory import CTreeFactory
from .models import Vendor

__all__ = ("CTreeSerializer",)


class CTreeSerializer:
    @classmethod
    def to_dict(cls, root: CTree) -> dict[str, Any]:
        children: dict[str, dict[str, Any]] = {}
        result = {
            "line": root.line,
            "tags": root.tags,
        }
        for child in root.children.values():
            children |= {child.line: cls.to_dict(child)}
        return result | {"children": children}

    @classmethod
    def from_dict(cls, vendor: Vendor, data: dict[str, Any], parent: CTree | None = None) -> CTree:
        _ct_class = CTreeFactory.get_class(vendor)
        node = _ct_class(line=data.get("line", ""), parent=parent, tags=data.get("tags", []))
        for child in data.get("children", {}).values():
            cls.from_dict(vendor, child, node)
        return node



from .abstract import CTree
from .postproc import CTreePostProc, register_rule
from .vendors import CiscoCT

__all__ = ("CiscoPostProcBGP",)


@register_rule
class CiscoPostProcBGP(CTreePostProc):
    @classmethod
    def _delete_nodes(cls, ct: CTree, lines_to_delete: set) -> None:
        nodes_to_delete: list[CTree] = []
        for node in ct.children.values():
            if len(node.children) != 0:
                cls._delete_nodes(node, lines_to_delete)
                if len(node.children) == 0:
                    nodes_to_delete.append(node)
            else:
                if node.line.startswith(tuple(lines_to_delete)) and not node.line.endswith(" peer-group"):
                    nodes_to_delete.append(node)
        for node in nodes_to_delete:
            node.delete()

    @classmethod
    def process(cls, ct: CTree) -> None:
        """Пост-обработка секции bgp для Cisco.

        - если есть команда "no neighbor <GROUP-NAME> peer-group", значит группы в целевой
            конфигурации вообще нет, и нужно все команды "no neighbor <GROUP-NAME> <some-options>"
            удалить, а оставить только одну "no neighbor <GROUP-NAME> peer-group", устройство
            остальное само вычистит

        Args:
            ct (CTree): дерево, для модификации
        """
        # на всякий случай проверим, что это дерево для cisco
        if not isinstance(ct, CiscoCT):
            return

        # найдем секцию bgp, если она есть
        bgp_nodes = [node for node in ct.children.values() if node.line.startswith("router bgp ")]
        if len(bgp_nodes) != 1:
            return
        bgp = bgp_nodes[0]

        # пересортируем секцию bgp, что бы все глобальные bgp команды шли перед настройками af
        bgp_global = {node.line: node for node in bgp.children.values() if len(node.children) == 0}
        bgp_af = {node.line: node for node in bgp.children.values() if len(node.children) != 0}
        bgp.children = bgp_global | bgp_af

        # теперь нужно сформировать признаки, как найти команды, которые будем удалять из diff'а
        lines_to_delete = set()
        for node in bgp.children.values():
            # можно через re поймать строку, но выбрал такой вариант
            if node.line.startswith("no neighbor ") and node.line.endswith(" peer-group"):
                # no neighbor <GROUP-NAME> peer-group
                _, _, group_name, _ = node.line.split()
                lines_to_delete.add(f"no neighbor {group_name}")

        # после получения признаков, удаляем ноды, которые им соответствуют
        cls._delete_nodes(bgp, lines_to_delete)



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from collections import deque

__all__ = ("CTree",)


class CTree(ABC):
    __slots__ = ["line", "parent", "children", "tags"]

    @property
    @abstractmethod
    def spaces(self) -> str:
        """количество пробелов для нового уровня."""

    @property
    @abstractmethod
    def section_exit(self) -> str:
        """как выходим из секции: exit/quit/..."""

    @property
    @abstractmethod
    def section_separator(self) -> str:
        """Чем разделяем блоки конфига между собой: !/#/..."""

    @property
    @abstractmethod
    def sections_without_exit(self) -> list[str]:
        """Список секций, из которых не нужно выходить.

        Некоторые секции не работают если явно выходить из них,
        например сертификаты в cisco, xpl в huawei.
        """

    @property
    @abstractmethod
    def sections_require_exit(self) -> list[str]:
        """Секции, которые могут быть пустыми.

        И для них необходимо явно прописывать выход,
        например route-map rm_DENY deny 10.
        """

    @property
    @abstractmethod
    def junk_lines(self) -> list[str]:
        """список линий, которые нужно игнорировать при анализе конфигурации."""

    @property
    @abstractmethod
    def undo(self) -> str:
        """как убираем конфигурацию, в общем случае: no/undo/..."""

    @property
    @abstractmethod
    def mask_patterns(self) -> list[str]:
        """паттерны для маскирования строк, указываем текст перед тем, что нужно заменить."""

    masking_string: str = "******"
    # empty_section_placeholder = "<-empty-section->"

    def __init__(self, line: str = "", parent: CTree | None = None, tags: list[str] | None = None) -> None:
        self.line = line.strip()

        # pattern = "|".join(self.mask_lines)
        # self.masked_line = re.sub(rf"({pattern}) \S+", rf"\1 {self.mask_pattern}", self.line)

        self.parent = parent
        self.children: dict[str, CTree] = {}

        if tags is not None:
            self.tags = tags
        elif parent is not None:
            self.tags = parent.tags
        else:
            self.tags = []

        if parent is not None:
            parent.children[line.strip()] = self

    @classmethod
    def mask_line(cls, line: str) -> str:
        # if len(cls.mask_patterns) == 0:
        #     return line
        pattern = "|".join(cls.mask_patterns)  # type: ignore [arg-type]
        if (m := re.fullmatch(pattern, line)) is not None:
            secret = [g for g in m.groups() if g is not None][0]
            return line.replace(secret, cls.masking_string)
        else:
            return line
        # return re.sub(rf"({pattern}) \S+", rf"\1 {cls.hidden_chars}", line)
        # return re.sub(pattern, rf"\g<before>{cls.masking_string}\g<after>", line)

    @property
    def masked_line(self) -> str:
        return self.mask_line(self.line)

    def __str__(self) -> str:
        """строковое представление: сама строка или 'root', если для корня вызываем."""
        return self.line or "root"

    def __repr__(self) -> str:
        """формальное представление объекта."""
        return f"({id(self)}) '{str(self)}'"

    def delete(self) -> None:
        to_delete = [self]
        stack = deque(self.children.values())
        while len(stack) > 0:
            node = stack.popleft()
            to_delete.append(node)
            if len(node.children) != 0:
                stack.extendleft(list(node.children.values())[::-1])
        for node in to_delete[::-1]:
            if node.parent is not None:
                _ = node.parent.children.pop(node.line)
                del node

    def __hash__(self) -> int:
        """вычисление hash."""
        return hash(self.formal_path)

    # todo добавить сравнение, с учетом порядка команд, как в differ сделано
    def __eq__(self, other: object) -> bool:
        """сравниваем два объекта.

        считаем узлы равными, если у них равны между собой:
        - строки настройки
        - одинаковые родители
        - одинаковые потомки

        Args:
            other (object): с чем сравниваем

        Returns:
            bool: равны или нет узлы
        """
        if not isinstance(other, CTree):
            return NotImplemented
        # возможно стоит сравнивать маскированные строки, что бы исключить разницу из-за хешей
        if self.line != other.line:
            return False
        if len(self.children) != len(other.children):
            return False

        self_parents = self._formal_path
        other_parents = other._formal_path

        if self_parents != other_parents:
            return False

        if set(self.tags) != set(other.tags):
            return False

        children_eq = []
        for line, node in self.children.items():
            other_node = other.children.get(line)
            if other_node is None:
                return False
            children_eq.append(node == other_node)
        return all(children_eq)

    @property
    def _formal_path(self) -> list[str]:
        result = []
        node = self
        while node.parent is not None:
            result.append(node.line)
            node = node.parent
        result.reverse()
        return result

    @property
    def formal_path(self) -> str:
        return " / ".join(self._formal_path)

    def _config(self, symbol: str, level: int, masked: bool) -> list[str]:
        line = self.masked_line if masked else self.line
        # if line == self.empty_section_placeholder:
        #     return []
        result = [symbol * level + line]
        for child in self.children.values():
            result.extend(child._config(symbol=symbol, level=level + 1, masked=masked))
        return result

    def _build_config(self, masked: bool) -> str:
        result = []
        level = 0
        node = self
        while node.parent is not None:
            result.append(node.masked_line if masked else node.line)
            level += 1
            node = node.parent
        result.reverse()
        result = [self.spaces * indx + line for indx, line in enumerate(result)]
        for child in self.children.values():
            result.extend(child._config(symbol=self.spaces, level=level, masked=masked))
            if self.parent is None:
                result.append(self.section_separator)
        return "\n".join(result)

    @property
    def config(self) -> str:
        return self._build_config(masked=False)

    @property
    def masked_config(self) -> str:
        return self._build_config(masked=True)

    @property
    def _formal_config(self) -> list[list[str]]:
        result = []
        for node in self.children.values():
            if len(node.children) == 0:
                result.append(node._formal_path)
            else:
                result.extend(node._formal_config)
        return result

    @property
    def formal_config(self) -> str:
        result = []

        node = self
        if len(self.children) == 0:
            return self.formal_path
        for node in self.children.values():
            if len(node.children) == 0:
                result.append(node._formal_path)
            else:
                result.extend(node._formal_config)

        return "\n".join([" / ".join(config) for config in result])

    def _build_patch(self, masked: bool) -> str:
        nodes = deque(self.children.values())
        result = []
        path_to_root = []

        node = self
        while node.parent is not None:
            path_to_root.append(node.masked_line if masked else node.line)
            node = node.parent
        path_to_root.reverse()

        while len(nodes) > 0:
            node = nodes.popleft()
            # if node.line == node.empty_section_placeholder:
            #     continue
            result.append(node.masked_line if masked else node.line)
            if len(node.children) != 0:
                if not re.fullmatch("|".join(self.sections_without_exit), node.formal_path):
                    nodes.appendleft(self.__class__(line=self.section_exit))
                nodes.extendleft(list(node.children.values())[::-1])
            elif len(self.sections_require_exit) != 0 and re.fullmatch(
                "|".join(self.sections_require_exit), node.formal_path
            ):
                nodes.appendleft(self.__class__(line=self.section_exit))
        result = path_to_root + result + [self.section_exit] * len(path_to_root)
        return "\n".join(result)

    @property
    def patch(self) -> str:
        return self._build_patch(masked=False)

    @property
    def masked_patch(self) -> str:
        return self._build_patch(masked=True)

    def _copy(self, children: bool, parent: CTree | None) -> CTree:
        if self.parent is not None and parent is None:
            parent = self.parent._copy(children=False, parent=None)

        new_obj = self.__class__(line=self.line, parent=parent, tags=self.tags.copy())
        if children:
            for child in self.children.values():
                _ = child._copy(children, new_obj)
        return new_obj

    def copy(self, children: bool = True) -> CTree:
        root = self._copy(children=children, parent=None)
        while root.parent is not None:
            root = root.parent
        return root

    def merge(self, other: CTree) -> None:
        for line, node in other.children.items():
            if line not in self.children:
                _ = node._copy(children=True, parent=self)
            else:
                self.children[line].merge(node)

    def _subtract(self, other: CTree, masked: bool = False) -> None:
        nodes_to_delete = []
        for child in self.children.values():
            line = child.exists_in(other, masked)
            if len(line) != 0:
                if len(child.children) != 0:
                    child._subtract(other.children[line])
                if len(child.children) == 0:
                    nodes_to_delete.append(child)
        for node in nodes_to_delete:
            node.delete()

    def subtract(self, other: CTree) -> CTree:
        result = self.copy()
        result._subtract(other=other)
        return result

    def _apply(self, other: CTree) -> None:
        for child in other.children.values():
            if child.line.startswith(child.undo):
                line = child.line.replace(child.undo, "").strip()
                if line in self.children:
                    self.children[line].delete()
            else:
                if child.exists_in(self):
                    self.children[child.line]._apply(other=child)
                else:
                    child._copy(children=True, parent=self)

    def apply(self, other: CTree) -> CTree:
        result = self.copy()
        result._apply(other=other)
        return result

    def rebuild(self, deep: bool = False) -> None:
        new_children = {child.line: child for child in self.children.values()}
        self.children = new_children
        if deep:
            for child in self.children.values():
                child.rebuild(deep)

    def exists_in(self, other: CTree, masked: bool = False) -> str:
        if masked:
            for line, node in other.children.items():
                if self.masked_line == node.masked_line:
                    return line
            return ""
        else:
            if self.line in other.children:
                return self.line
            else:
                return ""

    def reorder(self, tags: list[str], *, reverse: bool = False) -> None:
        def _get_children_tags(node: CTree) -> list[str]:
            tags = node.tags.copy()
            for child in node.children.values():
                tags.extend(_get_children_tags(child))
            return list(set(tags))

        if len(tags) == 0:
            return

        no_tags = "_no_tags_nodes"
        children: dict[str, list[CTree]]

        if not reverse:
            children = {tag: [] for tag in tags}
        else:
            children = {tag: [] for tag in reversed(tags)}
        children[no_tags] = []

        for child in self.children.values():
            child_tags = _get_children_tags(child)
            common_tags = set(tags).intersection(set(child_tags))
            if len(common_tags) == 0:
                children[no_tags].append(child)
            else:
                tag = common_tags.pop()
                children[tag].append(child)

        self.children = {}
        for child_list in children.values():
            for child in child_list:
                self.children[child.line] = child

    @classmethod
    def pre_run(cls, config: str) -> str:
        return config

    def post_run(self) -> None:
        return



from .abstract import CTree
from .postproc import CTreePostProc, register_rule
from .vendors import HuaweiCT

__all__ = (
    "HuaweiPostProcAAA",
    "HuaweiPostProcBGP",
    "HuaweiPostProcInterface",
    "HuaweiPostProcPrefixList",
    "HuaweiPostProcRoutePolicy",
    "HuaweiPostProcTacacs",
)


@register_rule
class HuaweiPostProcAAA(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        """Пост-обработка секции aaa.

        - Для пустых секций добавляем заглушку, что бы корректно генерировался патч
        - Работаем с пользователями:
            - если пароль пустой в diff, значит он не меняется и этот узел и
                соответствующий ему undo нужно удалить
            - если пароль пустой (не задан в vars) и при этом нет undo узла, значит сейчас
                на устройстве нет пользователя и нужно уделить все узлы с его именем, иначе
                система будет выпадать с ошибкой
            - если в diff есть пароль, значит мы его меняем, и соответствующий undo
                узел нужно удалить (только его)
            - если есть undo пользователя, тогда нужно удалить все ноды с его настройками
                и оставить только undo username
            - правим настройки, типа privilege level и пр, что бы они системой принимались
        - Если после всех манипуляций секция aaa стала пустой (а такое может быть, если diff
            состоял только из узлов с паролями, которые были удалены алгоритмом), тогда и саму
            секцию aaa нужно удалить
        """
        if not isinstance(ct, HuaweiCT):
            return
        aaa = ct.children.get("aaa")
        if not isinstance(aaa, HuaweiCT):
            return

        lines_to_delete = []
        nodes_to_delete = set()
        for node in aaa.children.values():
            ### шаг 1
            # пустой пароль - значит во входных данных ничего не было и пароль не меняем, поэтому
            # эти узлы, и соответствующие им undo-узлы нужно удалить
            if node.line.endswith("password irreversible-cipher"):
                _, user, *_ = node.line.split()
                lines_to_delete.append(f"local-user {user} password irreversible-cipher")
                undo_line = f"undo local-user {user} password irreversible-cipher"
                lines_to_delete.append(undo_line)
                # проверяем, если ли undo нода, если есть, значит пользователь существует уже в системе
                # если нет - значит нужно удалить все ноды для настройки этого пользователя
                if len([node for node in aaa.children.values() if node.line.startswith(undo_line)]) == 0:
                    lines_to_delete.append(f"local-user {user}")
            ### шаг 2
            # undo-узел и обычный узел с password для одного и того же пользователя,
            # значит мы меняем пароль, поэтому нужно undo-узел удалить
            elif not node.line.startswith("undo ") and " password irreversible-cipher " in node.line:
                _, user, *_ = node.line.split()
                lines_to_delete.append(f"undo local-user {user} password irreversible-cipher")

        nodes_to_delete.update([node for node in aaa.children.values() if node.line.startswith(tuple(lines_to_delete))])
        for node in nodes_to_delete:
            node.delete()
        aaa.rebuild()

        ### шаг 3
        # теперь, если остались undo-узлы с password, значит удаляем пользователя целиком, а значит
        # нужно поменять строку на undo local-user {user}, и удалить остальные узлы с этим пользователем
        lines_to_delete = []
        nodes_to_delete = set()
        for node in aaa.children.values():
            if node.line.startswith("undo ") and " password irreversible-cipher " in node.line:
                _, _, user, *_ = node.line.split()
                lines_to_delete.append(f"undo local-user {user} ")  # с пробелом в конце
                node.line = f"undo local-user {user}"  # а тут без, что бы endswith не захватил эту строку
        nodes_to_delete.update([node for node in aaa.children.values() if node.line.startswith(tuple(lines_to_delete))])
        for node in nodes_to_delete:
            node.delete()
        aaa.rebuild()

        ### шаг 4
        # правим undo узлы, что бы они принимались устройством
        for child in aaa.children.values():
            if not child.line.startswith("undo "):
                continue
            # видимо придется вносить еще или переделывать эту часть, пока только о level и service-type знаю
            if " service-type " in child.line or " level " in child.line:
                child.line = " ".join(child.line.split()[:4])
        aaa.rebuild()
        if len(aaa.children) == 0:
            aaa.delete()


@register_rule
class HuaweiPostProcBGP(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, HuaweiCT):
            return
        filtered_bgp = [node for node in ct.children.values() if node.line.startswith("bgp")]
        if len(filtered_bgp) != 1:
            return
        else:
            bgp = filtered_bgp[0]
        bgp_global = {node.line: node for node in bgp.children.values() if len(node.children) == 0}
        bgp_af = {node.line: node for node in bgp.children.values() if len(node.children) != 0}
        bgp.children = bgp_global | bgp_af
        bgp.rebuild()


@register_rule
class HuaweiPostProcInterface(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, HuaweiCT):
            return

        for child in ct.children.values():
            if child.line.startswith("undo interface "):
                child.line = child.line.replace("undo interface ", "clear configuration interface ")
            elif child.line.startswith("interface "):
                nodes_to_delete = set()
                for sub_child in child.children.values():
                    if sub_child.line == "undo port link-type hybrid":
                        sub_child.line = "undo port link-type"
                        nodes_to_delete.update(
                            [node for node in child.children.values() if node.line.startswith("undo port hybrid ")]
                        )
                for node in nodes_to_delete:
                    node.delete()
                child.rebuild()

        ct.rebuild()


@register_rule
class HuaweiPostProcPrefixList(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, HuaweiCT):
            return
        pl_statements: dict[str, list[str]] = {}
        to_delete: list[CTree] = []
        for child in ct.children.values():
            if child.line.startswith("ip ip-prefix "):
                _, _, pl_name, _, pl_indx, *_ = child.line.split()
                if pl_name not in pl_statements:
                    pl_statements[pl_name] = []
                pl_statements[pl_name].append(pl_indx)
        for child in ct.children.values():
            if child.line.startswith("undo ip ip-prefix "):
                _, _, _, pl_name, _, pl_indx, *_ = child.line.split()
                if pl_indx in pl_statements.get(pl_name, []):
                    to_delete.append(child)
                else:
                    child.line = f"undo ip ip-prefix {pl_name} index {pl_indx}"
        for node in to_delete:
            node.delete()
        ct.rebuild()


@register_rule
class HuaweiPostProcRoutePolicy(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, HuaweiCT):
            return
        for child in ct.children.values():
            if child.line.startswith("undo route-policy "):
                child.line = child.line.replace("permit ", "")
                child.line = child.line.replace("deny ", "")
        ct.rebuild()


@register_rule
class HuaweiPostProcTacacs(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, HuaweiCT):
            return
        filtered_tacacs = [
            node
            for node in ct.children.values()
            if node.line.startswith(("hwtacacs-server template", "hwtacacs server template"))
        ]
        if len(filtered_tacacs) != 1:
            return
        else:
            tacacs = filtered_tacacs[0]

        lines_to_delete = []
        nodes_to_delete = set()
        for node in tacacs.children.values():
            ### шаг 1
            # пустой пароль - значит во входных данных ничего не было и пароль не меняем, поэтому
            # этот узел, и соответствующий ему undo-узел нужно удалить
            # ("shared-key cipher  secondary", "shared-key cipher  third")
            if node.line.endswith("shared-key cipher"):
                lines_to_delete.append(node.line)
                lines_to_delete.append(f"undo {node.line}")
            ### шаг 2
            # undo-узел и обычный узел с password, значит мы меняем пароль, нужно undo-узел удалить
            elif not node.line.startswith("undo ") and " shared-key cipher " in node.line:
                lines_to_delete.append(f"undo {node.line}")
        nodes_to_delete.update(
            [node for node in tacacs.children.values() if node.line.startswith(tuple(lines_to_delete))]
        )
        for node in nodes_to_delete:
            node.delete()
        tacacs.rebuild()

        ### шаг 3
        # теперь, если остались undo-узлы с password, значит их нужно поменять,
        # отрезав пароль, иначе система не примет команду
        lines_to_delete = []
        nodes_to_delete = set()
        for node in tacacs.children.values():
            if node.line.startswith("undo ") and " shared-key cipher " in node.line:
                lines_to_delete.append(" ".join(node.line.split()[:-2]))
        nodes_to_delete.update(
            [node for node in tacacs.children.values() if node.line.startswith(tuple(lines_to_delete))]
        )
        for node in nodes_to_delete:
            node.delete()
        tacacs.rebuild()
        if len(tacacs.children) == 0:
            tacacs.delete()



import re
from typing import Type

from .abstract import CTree
from .postproc import _REGISTRY, CTreePostProc

__all__ = ("CTreeDiffer",)


class CTreeDiffer:
    @classmethod
    def _check_ordered(cls, a: CTree, ordered_sections: list[str] | None = None) -> bool:
        if ordered_sections is not None:
            for section in ordered_sections:
                formal_line = " / ".join(a._formal_path)
                if re.search(section, formal_line):
                    return True
        return False

    @classmethod
    def _check_no_diff(cls, a: CTree, no_diff_sections: list[str] | None = None) -> bool:
        if no_diff_sections is not None:
            for section in no_diff_sections:
                formal_line = " / ".join(a._formal_path)
                if re.search(section, formal_line):
                    return True
        return False

    @classmethod
    def _diff_list(
        cls,
        a: CTree,  # текущая конфигурация
        b: CTree,  # целевая
        *,
        existed_diff: CTree | None = None,
        ordered_sections: list[str] | None = None,
        no_diff_sections: list[str] | None = None,
        masked: bool = False,
        negative: bool = False,  # если True, то вычисляем, что нужно удалить, т.е. чего нет в целевой конфигурации
    ) -> list[CTree]:
        result = []
        _ordered = cls._check_ordered(a, ordered_sections)
        indx = 0
        if existed_diff is not None:
            b = b.apply(existed_diff)
        for child in a.children.values():
            # для секций, требующих полной перезаписи (без вычисления diff'a)
            _no_diff = cls._check_no_diff(child, no_diff_sections)
            if _no_diff:
                # делаем <undo> <section> (если она есть) когда negative=True
                #! upd: не делаем, потому что будет reordering и <undo> уедет в конец, если
                #! нужно удалять секцию, то это нужно добавлять через post-processing
                # if negative:
                #     if child.exists_in(b):
                #         root = child.copy(children=not negative)
                #         while len(node.children) == 1:
                #             node = list(node.children.values())[0]
                #         node.line = f"{node.undo} {node.line}"
                #         root.rebuild(deep=True)
                #         result.append(root)
                # целиком добавляем (negative=False)
                if not negative:
                    line = child.exists_in(b)
                    if len(line) == 0 or child != b.children.get(line):
                        root = child.copy(children=True)
                        result.append(root)
                continue
            if _ordered:
                if len(b.children) > indx and child.line == list(b.children.values())[indx].line:
                    line = child.line
                    indx += 1
                else:
                    line = ""
            else:
                line = child.exists_in(b, masked)
            if len(line) == 0:
                root = child.copy(children=not negative)
                if negative:
                    node = root
                    while len(node.children) == 1:
                        node = list(node.children.values())[0]
                    # добавить default? нужно кейс вспомнить
                    # _ = node.__class__(f"default {node.line}", node.parent, node.tags.copy())
                    if node.line.startswith(f"{node.undo} "):
                        node.line = node.line.replace(f"{node.undo} ", "", 1)
                    else:
                        node.line = f"{node.undo} {node.line}"
                    root.rebuild(deep=True)
                result.append(root)
            else:
                # todo тут если потомков нет, то нет смысла делать рекурсию, лишняя трата ресурсов
                nested_result = cls._diff_list(
                    child,
                    b.children[line],
                    existed_diff=None,
                    ordered_sections=ordered_sections,
                    no_diff_sections=no_diff_sections,
                    masked=masked,
                    negative=negative,
                )
                result.extend(nested_result)
        return result

    @classmethod
    def diff(
        cls,
        a: CTree,
        b: CTree,
        *,
        masked: bool = False,
        ordered_sections: list[str] | None = None,
        no_diff_sections: list[str] | None = None,
        reorder_root: bool = True,
        post_proc_rules: list[Type[CTreePostProc]] | None = None,
    ) -> CTree:
        # TODO тут подумать, что бы сразу в нужный parent крепить узел, а не делать merge списка потом

        if a.__class__ != b.__class__:
            raise RuntimeError("a and b should be instances of the same class")

        root = a.__class__()

        diff_list = cls._diff_list(
            a,
            b,
            existed_diff=None,
            ordered_sections=ordered_sections,
            no_diff_sections=no_diff_sections,
            masked=masked,
            negative=True,
        )
        for leaf in diff_list:
            root.merge(leaf)

        diff_list = cls._diff_list(
            b,
            a,
            existed_diff=root,
            ordered_sections=ordered_sections,
            no_diff_sections=no_diff_sections,
            masked=masked,
            negative=False,
        )
        for leaf in diff_list:
            root.merge(leaf)

        negative = {node.line: node for node in root.children.values() if node.line.startswith(node.undo)}
        for node in negative.values():
            node.tags.append("clear")
        if reorder_root:
            positive = {node.line: node for node in root.children.values() if not node.line.startswith(node.undo)}
            root.children = positive | negative

        if post_proc_rules is None:
            post_proc_rules = list(_REGISTRY.values())
        if isinstance(post_proc_rules, list):
            for rule in post_proc_rules:
                rule.process(root)

        return root



from .abstract import CTree
from .postproc import CTreePostProc, register_rule
from .vendors import AristaCT

__all__ = (
    "AristaPostProcAAA",
    "AristaPostProcBGP",
    "AristaPostProcPrefixList",
    "AristaPostProcTacacsKey",
    "AristaPostProcUsers",
)


@register_rule
class AristaPostProcAAA(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, AristaCT):
            return

        lines_to_check = (
            "aaa authentication login default",
            "aaa authentication login console",
            "aaa authentication enable default",
        )
        lines_to_delete = [
            f"no {' '.join(node.line.split()[:4])}"
            for node in ct.children.values()
            if node.line.startswith(lines_to_check)
        ]
        nodes_to_delete = set([node for node in ct.children.values() if node.line.startswith(tuple(lines_to_delete))])
        for node in nodes_to_delete:
            node.delete()
        ct.rebuild()


@register_rule
class AristaPostProcBGP(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        def _delete_nodes(ct: CTree) -> None:
            nodes_to_delete: list[CTree] = []
            for node in ct.children.values():
                if len(node.children) != 0:
                    _delete_nodes(node)
                    if len(node.children) == 0:
                        nodes_to_delete.append(node)
                else:
                    if node.line.startswith(tuple(lines_to_delete)) and not node.line.endswith(" peer group"):
                        nodes_to_delete.append(node)
            for node in nodes_to_delete:
                node.delete()

            ct.rebuild()

        if not isinstance(ct, AristaCT):
            return
        bgp_nodes = [node for node in ct.children.values() if node.line.startswith("router bgp ")]
        if len(bgp_nodes) != 1:
            return
        else:
            bgp = bgp_nodes[0]
        bgp_global = {node.line: node for node in bgp.children.values() if len(node.children) == 0}
        bgp_af = {node.line: node for node in bgp.children.values() if len(node.children) != 0}
        bgp.children = bgp_global | bgp_af
        bgp.rebuild()

        lines_to_delete = set()
        for node in bgp.children.values():
            if node.line.startswith("no neighbor ") and node.line.endswith(" peer group"):
                # no neighbor GROUP-PEER peer group
                _, _, peer_name, _, _ = node.line.split()
                lines_to_delete.add(f"no neighbor {peer_name}")
        _delete_nodes(bgp)
        bgp.rebuild()


@register_rule
class AristaPostProcPrefixList(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, AristaCT):
            return
        pl_statements: dict[str, list[str]] = {}
        to_change: list[CTree] = []
        # в pl_statements записываем pl и seq, которые будем настраивать
        for child in ct.children.values():
            if child.line.startswith("ip prefix-list "):
                _, _, pl_name, _, pl_indx, *_ = child.line.split()
                if pl_name not in pl_statements:
                    pl_statements[pl_name] = []
                pl_statements[pl_name].append(pl_indx)

        # проверяем, если no опция для комбинации pl+seq есть в pl_statements
        # значит мы меняем запись, если нет - значит просто удаляем
        for child in ct.children.values():
            if child.line.startswith("no ip prefix-list "):
                _, _, _, pl_name, _, pl_indx, *_ = child.line.split()
                if pl_indx in pl_statements.get(pl_name, []):
                    to_change.append(child)
                # и если просто удаляем, тогда меняем line у узла на тот формат, который принимается устройством
                else:
                    child.line = f"no ip prefix-list {pl_name} seq {pl_indx}"

        # те записи, которые меняем, нужно удалить перед тем, как настраивать
        clear_before_configure: dict[str, CTree] = {}
        for node in to_change:
            _, _, _, pl_name, _, pl_indx, *_ = node.line.split()
            new_node = node.__class__(
                line=f"no ip prefix-list {pl_name} seq {pl_indx}",
                parent=node.parent,
                tags=[tag for tag in node.tags if tag != "clear"],
            )
            clear_before_configure |= {new_node.line: new_node}
            node.delete()

        ct.children = clear_before_configure | ct.children
        ct.rebuild()


@register_rule
class AristaPostProcTacacsKey(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, AristaCT):
            return

        # если строка без пароля то удаляем этот и undo узлы
        if "tacacs-server key" in ct.children:
            lines_to_delete = ["tacacs-server key", "no tacacs-server key"]
        # если есть новый пароль, то удаляем только undo узел
        elif len([node for node in ct.children.values() if node.line.startswith("tacacs-server key ")]) == 1:
            lines_to_delete = ["no tacacs-server key"]
        else:
            lines_to_delete = []

        nodes_to_delete = set([node for node in ct.children.values() if node.line.startswith(tuple(lines_to_delete))])
        for node in nodes_to_delete:
            node.delete()
        ct.rebuild()


@register_rule
class AristaPostProcUsers(CTreePostProc):
    @classmethod
    def process(cls, ct: CTree) -> None:
        if not isinstance(ct, AristaCT):
            return

        lines_to_delete = []
        nodes_to_delete = set()
        for node in ct.children.values():
            if not node.line.startswith("username "):
                continue
            ### шаг 1
            # пустой пароль - значит во входных данных ничего не было и пароль не меняем, поэтому
            # эти узлы, и соответствующие им no-узлы нужно удалить
            if node.line.endswith("privilege 15 role network-admin secret"):
                _, user, *_ = node.line.split()
                lines_to_delete.append(f"username {user} privilege 15 role network-admin secret")
                lines_to_delete.append(f"no username {user} privilege 15 role network-admin secret sha512 ")
            ### шаг 2
            # обычный узел с password, значит мы меняем пароль, поэтому нужно no-узел удалить
            elif not node.line.startswith("no ") and " privilege 15 role network-admin secret " in node.line:
                _, user, *_ = node.line.split()
                lines_to_delete.append(f"no username {user} privilege 15 role network-admin secret sha512 ")
        nodes_to_delete.update([node for node in ct.children.values() if node.line.startswith(tuple(lines_to_delete))])
        for node in nodes_to_delete:
            node.delete()
        ct.rebuild()

        ### шаг 3
        # теперь, если остались no-узлы с password, значит удаляем пользователя целиком, а значит
        # нужно поменять строку на no username {user}
        lines_to_delete = []
        nodes_to_delete = set()
        for node in ct.children.values():
            if node.line.startswith("no ") and " privilege 15 role network-admin secret sha512 " in node.line:
                _, _, user, *_ = node.line.split()
                lines_to_delete.append(f"no username {user} ")  # с пробелом в конце
                node.line = f"no username {user}"  # а тут без, что бы endswith не захватил эту строку
        nodes_to_delete.update([node for node in ct.children.values() if node.line.startswith(tuple(lines_to_delete))])
        for node in nodes_to_delete:
            node.delete()
        ct.rebuild()



import re
from typing import Literal

from .abstract import CTree

__all__ = ("CTreeSearcher",)


class CTreeSearcher:
    @classmethod
    def _search(
        cls,
        ct: CTree,
        string: str,
        include_tags: list[str],
        include_mode: Literal["or", "and"],
        exclude_tags: list[str],
        include_children: bool,
    ) -> list[CTree]:
        """рекурсивный поиск."""
        result = []

        if len(string) == 0:
            string_match = True
        else:
            string_match = re.search(string, ct.line) is not None

        if len(include_tags) == 0:
            tags_match = True
        elif include_mode == "or" and not set(include_tags).isdisjoint(set(ct.tags)):
            tags_match = True
        elif include_mode == "and" and set(include_tags).issubset(set(ct.tags)):
            tags_match = True
        else:
            tags_match = False

        if not set(exclude_tags).isdisjoint(set(ct.tags)):
            tags_match = False

        match_result = all([string_match, tags_match])
        if match_result:
            result.append(ct.copy(children=include_children))
        if not match_result or not include_children:
            for child in ct.children.values():
                result.extend(
                    cls._search(
                        ct=child,
                        string=string,
                        include_tags=include_tags,
                        include_mode=include_mode,
                        exclude_tags=exclude_tags,
                        include_children=include_children,
                    )
                )

        return result

    @classmethod
    def search(
        cls,
        ct: CTree,
        *,
        string: str = "",
        include_tags: list[str] | None = None,
        include_mode: Literal["or", "and"] = "or",
        exclude_tags: list[str] | None = None,
        include_children: bool = False,
    ) -> CTree:
        """Поиск конфигурации в дереве.

        Args:
            ct (ConfigTree): где ищем
            string (str): что ищем, может быть regex строкой
            include_tags (list[str]): список тегов, по которым выборку делаем
            include_mode (Literal["or", "and"]): логика объединения критериев поиска
            exclude_tags (list[str]): список тегов-исключений, не должно быть на узле
            include_children (bool): включать потомков найденной секции или нет

        Returns:
            ConfigTree: новое дерево с отфильтрованным результатом
        """
        if include_tags is None:
            include_tags = []
        if exclude_tags is None:
            exclude_tags = []
        string = string.strip()
        root = ct.__class__()
        if len(string) == 0 and len(include_tags) == 0 and len(exclude_tags) == 0:
            return root
        filter_result = cls._search(
            ct=ct,
            string=string,
            include_tags=include_tags,
            include_mode=include_mode,
            exclude_tags=exclude_tags,
            include_children=include_children,
        )
        for node in filter_result:
            root.merge(node)
        return root



from __future__ import annotations

from typing import Type, cast

from .abstract import CTree
from .models import Vendor
from .vendors import AristaCT, CiscoCT, HuaweiCT

__all__ = ("CTreeFactory",)


class CTreeFactory:
    VENDOR_MAP = {
        Vendor.ARISTA: AristaCT,
        Vendor.CISCO: CiscoCT,
        Vendor.HUAWEI: HuaweiCT,
    }

    def __new__(cls, vendor: Vendor, line: str = "", parent: CTree | None = None) -> CTree:  # type: ignore[misc]
        _ct_class: Type[CTree] = cls.get_class(vendor)
        node = _ct_class(line=line, parent=parent)
        node = cast(CTree, node)
        return node

    @classmethod
    def get_class(cls, vendor: Vendor) -> Type[CTree]:
        _ct_class = cls.VENDOR_MAP.get(vendor)
        if _ct_class is None:
            raise NotImplementedError(f"unknown vendor {vendor}")
        return _ct_class



from .abstract import *
from .differ import *
from .environment import *
from .factory import *
from .models import *
from .parser import *
from .postproc import *
from .postproc_arista import *
from .postproc_cisco import *
from .postproc_huawei import *
from .searcher import *
from .serializer import *
from .vendors import *



from dataclasses import dataclass
from enum import StrEnum

__all__ = (
    "TaggingRule",
    "Vendor",
)


@dataclass(frozen=True, slots=True)
class TaggingRule:
    # - regex: ^ip vpn-instance (\\S+)$
    #   tags:
    #     - vpn
    #     - vrf
    # - regex: ^interface (\\S+)$
    #   tags:
    #     - interface
    regex: str
    tags: list[str]


class Vendor(StrEnum):
    ARISTA = "arista"
    CISCO = "cisco"
    HUAWEI = "huawei"



from pathlib import Path
from typing import Any, Literal

from .abstract import CTree
from .differ import CTreeDiffer
from .models import Vendor
from .parser import CTreeParser, TaggingRulesDict, TaggingRulesFile
from .postproc import CTreePostProc
from .searcher import CTreeSearcher
from .serializer import CTreeSerializer

__all__ = ("CTreeEnv",)


class CTreeEnv:
    def __init__(
        self,
        vendor: Vendor,
        *,
        tagging_rules: Path | str | list[dict[str, str | list[str]]] | None = None,
        ordered_sections: list[str] | None = None,
        no_diff_sections: list[str] | None = None,
        post_proc_rules: list[type[CTreePostProc]] | None = None,
    ):
        if isinstance(tagging_rules, str) or isinstance(tagging_rules, Path):
            _tr_file = TaggingRulesFile(tagging_rules)
        else:
            _tr_file = None
        if isinstance(tagging_rules, list):
            _tr_dict = TaggingRulesDict({vendor: tagging_rules})
        else:
            _tr_dict = None

        self.vendor = vendor
        self._parser = CTreeParser(vendor=self.vendor, tagging_rules=_tr_file or _tr_dict)
        self._ordered_sections = ordered_sections
        self._no_diff_sections = no_diff_sections
        self._post_proc_rules = post_proc_rules

    def parse(
        self,
        config: str,
    ) -> CTree:
        return self._parser.parse(
            config=config,
        )

    def diff(
        self,
        a: CTree,
        b: CTree,
        masked: bool = False,
        reorder_root: bool = True,
    ) -> CTree:
        return CTreeDiffer.diff(
            a=a,
            b=b,
            masked=masked,
            reorder_root=reorder_root,
            ordered_sections=self._ordered_sections,
            no_diff_sections=self._no_diff_sections,
            post_proc_rules=self._post_proc_rules,
        )

    def to_dict(
        self,
        ct: CTree,
    ) -> dict[str, Any]:
        return CTreeSerializer.to_dict(root=ct)

    def from_dict(
        self,
        data: dict[str, Any],
    ) -> CTree:
        return CTreeSerializer.from_dict(
            vendor=self.vendor,
            data=data,
        )

    def search(
        self,
        ct: CTree,
        *,
        string: str = "",
        include_tags: list[str] | None = None,
        include_mode: Literal["or", "and"] = "or",
        exclude_tags: list[str] | None = None,
        include_children: bool = False,
    ) -> CTree:
        return CTreeSearcher.search(
            ct=ct,
            string=string,
            include_tags=include_tags,
            include_mode=include_mode,
            exclude_tags=exclude_tags,
            include_children=include_children,
        )
